package Source::Package::Tarball::Python::Version;

use strict;
use warnings;
use Carp;

our $VERSION=0.002;
our $verbose=0;

# https://www.python.org/dev/peps/pep-0386/


#my $FRACTION_PART = qr/\.[0-9]+/;

# First part of either decimal or dotted-decimal strict version number.
# Unsigned integer with no leading zeroes (except for zero itself) to
# avoid confusion with octal.

#my $STRICT_INTEGER_PART = qr/0|[1-9][0-9]*/;

sub parse {
    my ($class,$versionstring)=@_;
    if ($versionstring =~ /^([0-9]+\.[0-9]+(?:\.[0-9]+)?)(?:([ab])([0-9]+))?$/ ) {
	my @args=(
	    -string=>$versionstring,
	    -mainversion=>&__split_on_dot($1),
	    -is_strict=>1,
	    -is_pep386_compatible =>1,
	    -is_pep386 => 1,
	    );
	push @args,
	    -pretype=>$2,
	    -preversion=>&__split_on_dot($3) if $2;
	return Source::Package::Tarball::Python::Version::PEP386->new(@args);

#expr = r"""^
#(?P<version>\d+\.\d+)         # minimum 'N.N'
#(?P<extraversion>(?:\.\d+)*)  # any number of extra '.N' segments
#(?:
#    (?P<prerel>[abc]|rc)         # 'a' = alpha, 'b' = beta
#                                 # 'c' or 'rc' = release candidate
#    (?P<prerelversion>\d+(?:\.\d+)*)
#)?
#(?P<postdev>(\.post(?P<post>\d+))?(\.dev(?P<dev>\d+))?)?
#$"""
   } elsif ($versionstring =~ /^
	    (\d+(?:\.\d+)*)
	    (?:
	    ([abc]|rc)
	    (\d+(?:\.\d+)*)
	    )?
	    (?:\.post(\d+))?(?:\.dev(\d+))?
	    $/x ) {
       my $mainver=&__split_on_dot($1);
       my $is_pep386=scalar @$mainver>1 ? 1:0;
	my @args=(
	   -is_pep386_compatible =>1,
	   -is_pep386 => $is_pep386,
	   -string=>$versionstring,
	   -mainversion=>$mainver,
	    );
       push @args, 
	   -pretype=>$2,
	   -preversion=>&__split_on_dot($3), if $2;
       push @args, 
	   -postversion=>&__split_on_dot($4), if $4;
       push @args, 
	   -devversion=>&__split_on_dot($5), if $5;
       return Source::Package::Tarball::Python::Version::PEP386->new(@args);
    } else {
	warn "Python::Version: version $versionstring is not strict, falling back to RPM\n" if $verbose;
	return Source::Package::Tarball::Python::Version::RPM->new(-string=>$versionstring);
    }
}

sub new {
    my $class=shift;
    my $self={
	@_,
    };
    bless $self,$class;
}

sub to_string {
    my $self=shift;
    return $self->{-string};
}

sub to_rpm_string {
    my $self=shift;
    return $self->to_string();
}

sub is_prerel {
    return $_[0]->{-pretype};
}

sub is_devel {
    return $_[0]->{-devversion};
}

use RPM::Vercmp;
sub cmp {
    my ($v1,$v2)=@_;
    if (ref($v1) eq 'Source::Package::Tarball::Python::Version::PEP386' and
	ref($v2) eq 'Source::Package::Tarball::Python::Version::PEP386') {
	#my $v1r=$v1->to_rpm_string();
	#my $v2r=$v2->to_rpm_string();
	#my $v1s=$v1->to_string();
	#my $v2s=$v2->to_string();
	#warn "PEP cmp: $v1r($v1s) $v2r($v2s)=",Source::Package::Tarball::Python::Version::PEP386::cmp($v1,$v2),"\n";
	return Source::Package::Tarball::Python::Version::PEP386::cmp($v1,$v2);
    } else {
	my $v1r=$v1->to_rpm_string();
	my $v2r=$v2->to_rpm_string();
	#if ($verbose>1) {
	    #use Data::Dumper;
	    #warn "RPM cmp: undef1: ",ref($v1),'->',Dumper($v1),"\n" unless defined $v1r;
	    #warn "RPM cmp: undef2: ",ref($v2),'->',Dumper($v2),"\n" unless defined $v2r;
	    #warn "RPM cmp: (",ref($v1),")$v1r (",ref($v2),")$v2r\n";
	#}
	return RPM::Vercmp::vercmp($v1r,0,$v2r,0);
    }
}

sub __split_on_dot {
    my $in=shift;
    #return undef unless defined ($_[0]);
    my @split=split(/\./, $in);
    #warn "split: debug: got $in returned [",join('][',@split),"]\n";
    return \@split;
}

sub __cmp_version_arrays {
    my ($v1,$v2)=@_;
    my $i=0;
    my $imax=$#{$v1};
    $imax=$#{$v2} if $#{$v2}>$imax;
    for ($i=0; $i<=$imax; $i++) {
	my $v1e=$v1->[$i] // 0;
	my $v2e=$v2->[$i] // 0;
	my $cmp = $v1e <=> $v2e;
	return $cmp if $cmp;
    }
    return 0;
}

package Source::Package::Tarball::Python::Version::PEP386;
use strict;
use warnings;
our @ISA=qw/Source::Package::Tarball::Python::Version/;

sub to_rpm_string {
    my $self=shift;
    my $ver=join('.',@{$self->{-mainversion}});
    my $post=$self->{-postversion};
    $ver.='.post'.join('.',@{$self->{-postversion}}) if $post;
    my $dev=$self->{-devversion};
    $ver.='.dev'.join('.',@{$self->{-devversion}}) if $dev;
    return $ver if !wantarray;
    
    my $rel;
    my $pretype=$self->{-pretype};
    $rel.=$pretype if $pretype;
    my $preversion=$self->{-preversion};
    $rel.=join('.',@{$preversion}) if $preversion;
    return $ver, $rel;
}

sub cmp {
    my ($v1,$v2)=@_;
    return 0 if $v1->to_string eq $v2->to_string;
    # main version
    my $v1main=$v1->{-mainversion};
    my $v2main=$v2->{-mainversion};
    my $cmp = Source::Package::Tarball::Python::Version::__cmp_version_arrays($v1main,$v2main);
    return $cmp if $cmp;
    # prerelease
    my $v1pretype=$v1->{-pretype};
    my $v2pretype=$v2->{-pretype};
    if ($v1pretype or $v2pretype) {
	return -1 if $v1pretype and !$v2pretype;
	return 1 if !$v1pretype and $v2pretype;
	$cmp = $v1pretype cmp $v2pretype;
	return $cmp if $cmp;
	my $v1preversion=$v1->{-preversion};
	my $v2preversion=$v2->{-preversion};
	$cmp = Source::Package::Tarball::Python::Version::__cmp_version_arrays($v1preversion,$v2preversion);
	return $cmp if $cmp;
    }
    # postrelease
    my $v1post=$v1->{-postversion};
    my $v2post=$v2->{-postversion};
    if ($v1post or $v2post) {
	return 1 if $v1post and !$v2post;
	return -1 if !$v1post and $v2post;
	$cmp = Source::Package::Tarball::Python::Version::__cmp_version_arrays($v1post,$v2post);
	return $cmp if $cmp;
    }
    # devrelease
    my $v1dev=$v1->{-devversion};
    my $v2dev=$v2->{-devversion};
    if ($v1dev or $v2dev) {
	return 1 if $v1dev and !$v2dev;
	return -1 if !$v1dev and $v2dev;
	$cmp = Source::Package::Tarball::Python::Version::__cmp_version_arrays($v1dev,$v2dev);
	return $cmp if $cmp;
    }
    return 0;
}

package Source::Package::Tarball::Python::Version::RPM;
use strict;
use warnings;
our @ISA=qw/Source::Package::Tarball::Python::Version/;

package Source::Package::Tarball::Python::Version::LooseVersion;
use strict;
use warnings;
our @ISA=qw/Source::Package::Tarball::Python::Version/;

# TODO: see
# https://www.python.org/dev/peps/pep-0386/

1;


__END__

